with Ada.Text_IO; --  for access to basic I/O functions

package body SoE is

   --  each instance of process Sieve gets an own Id
   --+ (realized as a static variable at library level)
   Assigned_Id : Natural := 0;

   --  slow down factor for the activity of process Odd
   Slow_Down_Factor : constant Duration := 0.5;

   --+-----------------------
   protected body Buffer_T is
      --  values can be written into A_Container
      --+ only if that is not full
      --+ hence Put is a potentially blocking operation
      entry Put (Item : Integer) when Contents < Limit is
      begin
         A_Container (Insert) := Item;
         Insert   := Insert + 1;
         Contents := Contents + 1;
      end Put;
      --+--------------------
      --  values can be fetched from A_Container
      --+ only as long as that is not empty
      --+ (Get thus is a potentially blocking operation)
      entry Get (Item : out Integer) when Contents > 0 is
      begin
         Item     := A_Container (Extract);
         Extract  := Extract + 1;
         Contents := Contents - 1;
      end Get;
      --+--------------------
   end Buffer_T;
   --+-----------------------

   --+-----------------------
   task body Odd is
      Num : Positive;
      --  process Odd will share a pointer (an access value) to
      --+ the Buf vector with the top-level instance of the
      --+ chain of processes of type Sieve_T
      --+ (this language wants all pointers that may be shared
      --+  to be declared as "aliased")
      Shared_Buf : aliased Buffer_T;
      --  process Odd will create a child process Sieve (instance of Sieve_T)
      Sieve       : Sieve_Ref_T;
   begin
      --  process S is now activated and starts running
      --+ concurrently to Odd
      --+-------------------------------------------------------------------
      --+ process S create Sieve and gives it access to its own Buf vector,
      --+ which provides protected access to the A_Container array
      Assigned_Id := Assigned_Id + 1;
      Sieve       := new Sieve_T (Shared_Buf'Unchecked_Access, Assigned_Id);
      --+-------------------------------------------------------------------
      Num := 3;
      while Num < Limit loop
         --  Odd fills A_Container with all odd numbers
         --+ from 3 until Limit,
         --+ the first of those numbers (3) is known to be a prime
         Shared_Buf.Put (Num);
         Num := Num + 2;
         --  Odd slows down to rest a little in between
         delay Slow_Down_Factor;
      end loop;
      --  process Odd will complete when it will have passed down
      --+ the last odd number to its Sieve child, and will await the
      --+ termination of it to be able to terminate itself
   end Odd;
   --+-----------------------

   --+-----------------------
   task body Sieve_T is
      use Ada.Text_IO;
      --  Inner_Shared_Buf has exactly the same role as Shared_Buf
      --+ for process Odd
      Inner_Shared_Buf : aliased Buffer_T;
      Sieve            : Sieve_Ref_T;
      Prime, Num       : Positive;
   begin
      --  Sieve fetches the first value from the array
      --+ it shares with its master via the protected buffer B,
      --+ which was passed to it as a discriminant
      --+ (this first value is known to be a prime)
      Buf.Get (Prime);
      loop
         --  Sieve tries the next odd number off B
         --+ until it finds one that is not divisible
         --+ by the one and only prime it knows
         Buf.Get (Num);
         if Num rem Prime /= 0 then
            exit;
         end if;
      end loop;
      Put_Line ("Sieve instance " & Natural'Image (Id) &
                " found prime number" & Natural'Image (Num));
      --  As we proceed incrementally, Num is a new prime,
      --+ hence the Sieve process needs to pass Num down
      --+ to a child instance for it to continue the search
      --+ on all successive odd numbers not divisible by Prime
      --+ (which might be primes)
      --+-------------------------------------------------------------------
      Assigned_Id := Assigned_Id + 1;
      Sieve       := new Sibling_T (Inner_Shared_Buf'Unchecked_Access,
                                    Assigned_Id);
      --+-------------------------------------------------------------------
      Inner_Shared_Buf.Put (Num);
      loop
         Buf.Get (Num);
         if Num rem Prime /= 0 then
            Inner_Shared_Buf.Put (Num);
         end if;
      end loop;
      --  what can we say about the termination of
      --+ the hierarchy of instance(s) of Sieve_T that
      --+ this code creates?
      --+
      --+ (to actively force self-termination of all tasks
      --+  down the hiearchy we could define a shut-down value
      --+  and pass it down to all tasks: how?)
   end Sieve_T;
   --+-----------------------

end SoE;
